!******************************************************************************
!
! File : hamfunc.f90
!   originally from hamiltonian.f90
!   by : Alan Tackett
!   on : 02/07/97
!  for : PAW
!
! Module for calculating the Hamiltonian and Cohesive Energy of the PAW
! system.  Also contains routines to do H*Psi and O*Psi.
!
!  Contains relatively simple functions in construction of Hamiltonian
!      terms 
!
! Updated for spin by Ping Tang and N. A. W. Holzwarth
!      Last change 5/25/05
!******************************************************************************

Module hamfunc

Use atom_data
Use crystal_symmetry
Use gausslib
Use gpoints 
Use mathlib
Use options_data
Use paw_inout

Implicit NONE!!!!

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  RestrictPot - Restricts the potential from the High planewave grid to the
!       low cutoff grid in FFT space.
!
!  Vin    - High PW grid in FFT space(Input)
!  Vout   - Low PW grid in FFT space(Returned) 
!
!******************************************************************************

Subroutine RestrictPot( Vin, Vout)
  Complex, Intent(IN) :: Vin(:)
  Complex, Intent(OUT) :: Vout(:)

  Integer :: i, N_low, N_high
  real    :: weight, Ec

  Ec = PW_Gcut(G_PROJ)*(1-V_Smooth_width)

  N_low = Gpnt_Size(G_PROJ) - 1
  N_High = Gpnt_Size(G_HIGH) - 1

  Vout = 0

Write(Log_Unit,*) 'RestrictPot: PW_Gcut=',PW_Gcut, ' * V_Smooth_Width=',V_Smooth_Width
  Vout(FFTmap_PROJ(1)) = Vin(FFTmap_High(1))
  Do i=2, Gpnt_size(G_PROJ)
!     If (V_Smooth_Width > 0.0) then
!        weight = abs(Gpnt(4,i)-PW_Gcut(G_PROJ))/(PW_Gcut(G_Proj)*V_Smooth_Width)
!        weight = 2.0/(exp(-weight) + 1.0) - 1
!     else
!        Weight = 1
!     End if

     Weight = 1
     If (Gpnt(4,i) > Ec) then
        Weight = Pi*(PW_Gcut(G_PROJ) - Gpnt(4,i)) / &
             (2*V_Smooth_Width*PW_Gcut(G_PROJ))
        Weight = sin(weight)**2
     End if
!Write(Log_Unit,*) 'RestrictPot: i=',i, ' * |G|=',Gpnt(4,i), ' * Wt=',weight 
!weight = 1
     Vout(FFTmap_HIGH(i)) = Vin(FFTmap_High(i)) * Weight
     Vout(FFTmap_HIGH(i+N_HIGH)) = Vin(FFTmap_High(i+N_High)) * Weight
  End Do

  Return
End Subroutine


!******************************************************************************
!                  
!  CalcVlocal - Calculates Vlocal                        
!                  
!******************************************************************************

Subroutine CalcVlocal(V)
  Complex,         Intent(INOUT) :: V(:)

  Real    :: theta, Gmag, RadG, G(3), Esum
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi
  Type (Atom_Info_Fixed), Pointer :: AT

  V = 0

  RadIndex = 0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     !*** If needed inc the RadIndex ***
     If (G_NewMag(gindex)) RadIndex = RadIndex + 1

     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
        RadG = AT%Rad_VLocal(RadIndex)

        !*** Create the structure factors and accumulate vlocal***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
        
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
!Phase = CONJG(Phase)
           Vsum = Vsum + Phase*RadG
        End Do
     End do

     V(gindex) = V(gindex) + Vsum
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  function spheresymFT
!
!******************************************************************************

Function spheresymFT(ngrid,radialfn,Gmag,dr)
  Real :: spheresymFT
  Integer, INTENT(IN) :: ngrid
  Real, INTENT(IN) :: radialfn(:),Gmag,dr
 
  Real, allocatable :: fn(:)
  Integer       :: i
  Real          :: r, con,x

   con = Four_Pi

   Allocate(fn(ngrid))

   do i=1,ngrid
      r=dr*(i-1)
      x=r
      if (Gmag > machine_precision) x=SIN(Gmag*r)/Gmag
      fn(i) = con*r*radialfn(i)*x
   End do
          
  spheresymFT=IntSimpson(ngrid,dr,fn)

  deallocate(fn)
End Function spheresymFT

Subroutine CalcCoreTail(V)
  Complex, INTENT(OUT) :: V(:)
  Real    :: theta, Gmag, RadG, G(3), Esum,dr
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi,ngrid,i,j
  Type (Atom_Info_Fixed), Pointer :: AT
 
  Write(Log_unit,*) 'In CalcCoreTail',size(V)
  Do atype=1, Atom_Types
     AT => AtomType_Info(atype)
     Allocate(AT%FTCoreTail(NewMag_Size(G_High)), STAT=j)
        if (j /= 0) then
           write(Error_unit,*) 'Error CalcCoreTail: allocation ', &
                    NewMag_Size(G_High),atype,j
           stop
        endif
     AT%FTCoreTail=0
     dr= AT%Mesh_Step
     ngrid=AT%CoreTail_Points
     Radindex=0
     Do gindex=1, Gpnt_Size(G_HIGH)
         If (G_NewMag(gindex)) then
            Gmag = Gpnt(4,gindex)
            RadG = spheresymFT(ngrid,AT%CoreTail_Density,Gmag,dr)
            RadIndex=RadIndex+1
            AT%FTCoreTail(RadIndex)=RadG
         Endif
     Enddo
  Enddo
    
  V=0; Radindex=0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     If (G_NewMag(gindex)) Radindex=Radindex+1
 
     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
                                                                                
        !*** Create the structure factors and accumulate vlocal***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
                                                                                
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
           Vsum = Vsum + Phase*AT%FTCoreTail(RadIndex)
        End Do
     End do
      
     V(gindex)=Vsum
  End Do
End Subroutine

                                                                                
Subroutine UpdateCoreTail(V)
  Complex, INTENT(OUT) :: V(:)
  Real    :: theta, Gmag, RadG, G(3), Esum,dr
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi,ngrid,i,j
  Type (Atom_Info_Fixed), Pointer :: AT
                                                                                
  Write(Log_unit,*) 'In UpdateCoreTail',size(V)
                                                                                
  V=0; Radindex=0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     If (G_NewMag(gindex)) Radindex=Radindex+1
                                                                                
     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
                                                                                
        !*** Create the structure factors and accumulate coretail***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
                                                                                
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
           Vsum = Vsum + Phase*AT%FTCoreTail(RadIndex)
        End Do
     End do
                                                                                
     V(gindex)=Vsum
  End Do
 End subroutine
                                                                                

!******************************************************************************
!
!  CalcRadial_Vlocal - Calculates the radial Vlocal grids
!
!******************************************************************************

Subroutine CalcRadial_Vlocal
 
  Integer       :: i,j, k, G, Gmag, n, L,ngrid
  Character*200 :: Token
  Real, Pointer :: RadV(:),fn(:)
  Real          :: dr, r,  norm,xmax,gmax,con
  Type (Atom_Info_Fixed), Pointer :: AT

   con = Four_Pi/Xtal%volume
   L=0  ! L=0 for Vlocal

   ngrid=MAXVAL(AtomType_Info(:)%Mesh_Size)
   Allocate(fn(ngrid))

  Do i=1, Atom_Types
    AT => AtomType_Info(i)

    Allocate(AT%Rad_Vlocal(NewMag_Size(G_High)), STAT=j)
    Write(Token,*) 'CalcRadial_Vlocal: Error allocating array. Size=',&
         NewMag_Size(G_HIGH), ' * TypeIndex=',i
    Call Check_Error(j, Token, Error_Unit, .TRUE., PAW_WC,'CalcRadial_Vlocal:')

    RadV => AT%Rad_Vlocal

    RadV = 0

    If (Atomic_Mode == SIM_SEPM) then 
       k = 0
       con = 1.0 /xtal%volume
       Do j=1, Gpnt_Size(G_HIGH)
          If (G_NewMag(j)) then
             k = k + 1
             RadV(k) = EvalGauss(AT%V_sepm, Gpnt(4,j)) * con
          End If
       End Do
    else
      ! norm = con*AT%V_Local
       norm = con
       
       if (ABS(norm).gt.1.d-9) then
          
          DR=AT%Mesh_Step      
          write(6,*) 'CalcRadial_Vlocal: DR',DR
          
          !*** Tabulate the vlocal function ***
          Do j=1, AT%Mesh_Size
             r = DR*(j-1)
             
             !Fn(j) = norm*r*AT%Shape_Func(j)
             Fn(j) = norm*r*AT%RadR_Vlocal(j)
          End do
          
          k = 0
          Do j=1, Gpnt_Size(G_HIGH)
             If (G_NewMag(j)) then
                k = k + 1
                !RadV(k) = RadialFourier(L, Gpnt(4,j), DR, AT%Mesh_Size, fn)
                RadV(k) = spheresymFT(AT%Mesh_Size,&
                    AT%RadR_Vlocal,Gpnt(4,j),dr)/xtal%volume
             End If
          End Do
       Endif
    End If
  End Do

  Deallocate(fn)

  Return
End Subroutine


!******************************************************************************
!
!  CalcRadialHat - Calculates the radial n^ grids
!
!******************************************************************************

Subroutine CalcRadialHat

  Integer       :: i,j, k, G, Gmag, n, L, ngrid
  Character*200 :: Token
  Real, Pointer :: RadHat(:),fn(:)
  Real          ::  dr, r,  norm,xmax,gmax
  Type (Atom_Info_Fixed), Pointer :: AT


   ngrid=MAXVAL(AtomType_Info(:)%Mesh_Size)
   Allocate(fn(ngrid))


  Do i=1, Atom_Types
    AT => AtomType_Info(i)

    AT%Hat_MaxL = 2*MaxVal(AT%L_Value)
  
    Allocate(AT%RadHat(NewMag_Size(G_High), AT%Hat_MaxL+1), STAT=j)
    Write(Token,*) 'CalcRadialHat: Error allocating array. Size=',&
         NewMag_Size(G_HIGH), ' * ', AT%Basis_Size, ' * TypeIndex=',i
    Call Check_Error(j, Token, Error_Unit, .TRUE., PAW_WC, 'CalcRadialHat:')
    
    DR=AT%Mesh_Step    
     !write(6,*) 'CalcRadHat: DR',DR

    Do L=0, AT%Hat_MaxL
       RadHat => AT%RadHat(:,L+1)

       n=AT%Mesh_Size
       do j=1,n            
         fn(j)=((dr*(j-1))**(2*L+2))*AT%Shape_Func(j)
       enddo

       norm = sqrt(Four_Pi)/IntSimpson(n, DR, fn)

       !*** Tabulate the hat function ***
       Do j=1, n
          r = DR*(j-1)
          Fn(j) = norm * AT%Shape_Func(j) * r**(L+1)
       End do

       k = 0
       Do j=1, Gpnt_Size(G_HIGH)
          If (G_NewMag(j)) then
             k = k + 1
             RadHat(k) = RadialFourier(L, Gpnt(4,j), DR, n, fn)
            !write(6,'("radhat",2i4,1p2e15.7)') k,L,Gpnt(4,j),radHat(k)
          End If
       End Do
    End Do

  End Do

  Deallocate(fn)

  Return
End Subroutine

!******************************************************************************
!
! CalcKineticEnergy - Calculates the KE for the given Psi nad K-pnt
!
!   BlochK - K-Point
!   Psi    - Wave function
!
!******************************************************************************
                                                                                                             
Real Function CalcKineticEnergy(BlochK, Psi)
  Real,    Intent(IN) :: BlochK(:)
  Complex, Intent(IN) :: Psi(:)
                                                                                                             
  Integer :: gi, N
  Real    :: G(3), G2
  Complex :: KE
                                                                                                             
                                                                                                             
  N = Gpnt_Size(G_LOW) - 1
                                                                                                             
  KE = 0
  Do gi=1, Gpnt_Size(G_LOW)
     G = BlochK + Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     KE = KE + CONJG(Psi(gi))*Psi(gi) * G2
  End Do
                                                                                                             
  Do gi=2, Gpnt_Size(G_LOW)
     G = BlochK - Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     KE = KE + CONJG(Psi(gi+N))*Psi(gi+N) * G2
  End Do
                                                                                                             
  CalcKineticEnergy = KE
                                                                                                             
  Return
End function

End Module
